// uncomment the line to run in PC test mode
//#define debug
//#define STDMON

#ifndef STDMON
#define SUPPORT_DBCS
#endif

#define MODEX_DELAY     6   // default modex delay..

#include <dos.h>
#include <stdlib.h>

#include "define.h"
#include "mode15k.h"
#include "drv15k.h"
#include "tsr.h"
#include "tsrsub.h"
#include "vesa.h"


#ifdef SUPPORT_DBCS

static void DBCS_get_vector(void);
BYTE dbcs_supported;
WORD dbcs_ds;
WORD dbcs_si;

#endif


/* Stack size */
extern unsigned _stklen=1024;


/* Function to call old interrupt 10h */
#define INT10 \
        asm { push bp }  \
        (*old_int10)(); \
        asm { pop bp }

/* Function to call old interrupt  8h */
#define INT8 \
        asm { push bp }  \
        (*old_int8)(); \
        asm { pop bp }

WORD CalcBytesPerLine(int width, int depth)
{
    return ((WORD)width * depth) / 8;
}

WORD CalcPixelsPerLine(int offset, int depth)
{
    return ((WORD)offset * 8) / depth;
}

void VsyncVga(void)
{
    while( (_inb(0x3da) & 0x08) != 0 );
    while( (_inb(0x3da) & 0x08) == 0 );
}

int ParseOption(char *option)
{
    int temp;
    int c;
    char *opt;

    opt = &option[1];           /* Pointer to actual option */

    /* Force uppercase if a to z */
    if(option[0]>=97 && option[0]<=122)
    {
        option[0]-=32;
    }

    switch ( option[0] )
    {
        case 'H':
        case '?':
                printf(usage);
                return 0;

        case 'U':
                PROGRAM.unload=1;
                break;

        case 'E':
                PROGRAM.enabled=1-PROGRAM.enabled;
                break;

        case 'M':
                temp=atoi(opt);
                if(temp>48)
                        printf("Invalid -m option\n");
                else
                        PROGRAM.vidmem=temp;
                break;

        case 'W':
                temp=atoi(opt);
                if(temp<0 || 255<temp)
                        printf("Invalid -w option\n");
                else
                        PROGRAM.max_ticks=temp;
                break;

        case 'X':
                temp=atoi(opt);
                if(temp<-16 || temp>16)
                        printf("Invalid -x option\n");
                else
                        PROGRAM.center_x=temp;
                break;

        case 'Y':
                temp=atoi(opt);
                if(temp<-8 || temp>8)
                        printf("Invalid -y option\n");
                else
                        PROGRAM.center_y=temp;
                break;

        case '1':
                temp=atoi(opt);
                if(temp<0 || temp>3)
                        printf("Invalid -1 option\n");
                else
                        PROGRAM.h_sync_delay=temp;
                break;

        case '2':
                temp=atoi(opt);
                if(temp<0 || temp>3)
                        printf("Invalid -2 option\n");
                else
                        PROGRAM.h_disp_delay=temp;
                break;

        case 'P':
                PROGRAM.use_pm_interface=1-PROGRAM.use_pm_interface;
                if(own_pm_interface_supported() == 0){
                        PROGRAM.use_pm_interface = 0;
                }
                break;
#ifdef SUPPORT_DBCS
        case 'D':
                PROGRAM.use_dbcs = 1 - PROGRAM.use_dbcs;
                break;
#endif
        default:
            printf("Unknown option %c\n",option[0]);
            return 0;
    }
    return 1;
}

int ParseOptions(int optCount, char **options)
{
        int i;

        for ( i = 0; i < optCount; i++ )
        {
                /* `-' and `/' valid operator seperators */
                if((options[i][0]=='-') || (options[i][0]=='/'))
                {
                        if(ParseOption(&options[i][1])==0)
                        {
                                return 0; /* Help shown don't do anything */
                        }
                }
                else
                {
                    printf("Invalid command line option\n");
                    return 0;
                }
        }

        /* No amount specified, use Videocard amount */
        if(PROGRAM.vidmem==0)
        {
               PROGRAM.vidmem=VESA.memory>>4;
        }
         /* Allow memory emulation */
        VESA.memory=((int)PROGRAM.vidmem<<4);
        PROGRAM.membytes=(unsigned long)(VESA.memory<<6)<<10;
        return 1;
}

void ShowOptions()
{
        /* Activate to the arcade monitor/PC monitor depending on state */
        _AX=3;
        asm int 10h

        /* Display current options */
        printf("Status:-\n");
        printf("\t enabled = %s\n", YesNo[PROGRAM.enabled]);

        printf("\t video memory = %d MB\n", PROGRAM.vidmem);
        printf("\t ModeX wait time = %d\n", PROGRAM.max_ticks);

        printf("\t center_x = %i\n", PROGRAM.center_x);
        printf("\t center_y = %i\n", PROGRAM.center_y);
        printf("\t horizontal sync delay = %i\n", PROGRAM.h_sync_delay);
        printf("\t horizontal disp delay = %i\n", PROGRAM.h_disp_delay);

        if(own_pm_interface_supported() != 0){
                printf("\t own protected mode interface = %s\n", YesNo[PROGRAM.use_pm_interface]);
        }

#ifdef SUPPORT_DBCS
        printf("\t DBCS 640x480 text mode = %s\n", YesNo[PROGRAM.use_dbcs]);
#endif
}

void main(int argc, char *argv[])
{
        unsigned int TSR_SEG;

#ifdef SUPPORT_DBCS
        DBCS_get_vector();
#endif

        if(DetectVESA()==0)    /* No VESA support */
                return;

        if(Detect15KHz()==0)     /* Add your video card check here */
        {
                printf("%s", NotFoundMessage());
                return;
        }

        old_int10 = getvect(0x10);

        /* Check to make sure not already loaded */
        if (FP_OFF(old_int10)==FP_OFF(int10h_proc))
        {
                /* Get the tsr segment, add correct offest in segment for old interrupt vectors */
                TSR_SEG=FP_SEG(old_int10);

                memcpy(&PROGRAM,MK_FP(TSR_SEG+(FP_SEG(&PROGRAM)-FP_SEG(int10h_proc)),FP_OFF(&PROGRAM)),sizeof(PROGRAM));

                /* Check to make sure we got the right memory location */
                if (PROGRAM.loaded_flag==TSR_ID)
                {
                        /* TSR already in Memory */

                        /* Override the options with any new ones! */
                        if(ParseOptions(argc-1, &argv[1])==0)
                        {
                                return; /* Help was shown don't do anything */
                        }

                        if(PROGRAM.unload==1) /* Unload TSR */
                        {
                                /* Copy old Interrupt addresses */
                                memcpy(&old_int10,MK_FP(TSR_SEG+(FP_SEG(&old_int10)-FP_SEG(int10h_proc)),FP_OFF(&old_int10)),4);
                                memcpy(&old_int8,MK_FP(TSR_SEG+(FP_SEG(&old_int8)-FP_SEG(int10h_proc)),FP_OFF(&old_int8)),4);

                                asm cli

                                if(old_int10!=NULL)
                                        setvect(0x10,old_int10);

                                if(old_int8!=NULL)
                                        setvect(0x8,old_int8);

                                asm sti

                                asm mov ax,3
                                asm int 10h

                                printf("Restored old Int 0x10 at (%p)\n",old_int10);
                                printf("Restored old Int  0x8 at (%p)\n",old_int8);

                                freetsr(PROGRAM.psp);
                        }
                        else  /* Change options */
                        {
                                memcpy(MK_FP(TSR_SEG+(FP_SEG(&PROGRAM)-FP_SEG(int10h_proc)),FP_OFF(&PROGRAM)),&PROGRAM,sizeof(PROGRAM));
                                ShowOptions();
                        }
                }
                else    /* TSR is at the same offset, but failed the TSR flag, could be ours or someone elses */
                {
                        printf("Unable to unloaded.\n");
                }
                return;
        }

        /* TSR not loaded, lets do it! */

        /* Set default options first, before calling parse cmdline options */
        PROGRAM.psp=_psp;
        PROGRAM.loaded_flag=TSR_ID;

        PROGRAM.current_ticks=0;
        PROGRAM.unload=0;
        PROGRAM.enabled=1;
        PROGRAM.vidmem=0;
        PROGRAM.max_ticks=MODEX_DELAY; /* 18 = 1 second approx */

        PROGRAM.center_x=0;
        PROGRAM.center_y=0;
        PROGRAM.h_sync_delay=0;
        PROGRAM.h_disp_delay=0;

        PROGRAM.use_pm_interface = (own_pm_interface_supported() == 0) ? 0 : 1;
#ifdef SUPPORT_DBCS
        PROGRAM.use_dbcs = 0;
#endif
        /* Any user defined options?? */
        if(ParseOptions(argc-1, &argv[1])==0)
        {
                return; /* Help was shown don't do anything */
        }

        /* Set new interrupt vectors */
        asm cli
        set_old10h_proc( (DWORD)old_int10 );
        setvect(0x10, int10h_proc);

        old_int8 = getvect(0x08);
        setvect(0x08,int_8);

        asm sti

        ShowOptions();

        // Terminate and Stay Resident
        tsr();
}

void tsr()
{
        _DX=(_SS + ((_SP+16)/16) - _psp);
        _AH=0x31;
        _AL=0;
        asm { int 21h };
}

void freetsr(unsigned int prog_seg)
{
        unsigned int seg;

        for ( seg = 0; ! IS_FIRST_ARENA_HEADER( seg ); seg++ );

        while ( ! IS_LAST_ARENA_HEADER( seg ) )
        {
                if (FAR_WORD(seg, 1) == prog_seg)
                {
                        /* DOS free segment */
                        _AX=seg+1;
                        _ES=_AX;
                        _AH=0x49;
                        asm     int 21h
                }
                seg = NEXT_ARENA_HEADER( seg );
        }
}

void interrupt int_8(void)
{
        asm     cli

        INT8

        if (PROGRAM.current_ticks) /* We have a possible pending modex mode */
        {
                if(PROGRAM.current_ticks==PROGRAM.max_ticks)
                {
                        PROGRAM.current_ticks=0;
                        SetVGA(640, 240, PROGRAM.center_x, PROGRAM.center_y, PROGRAM.h_sync_delay, PROGRAM.h_disp_delay);   // for modex 320x240
                        SetTextHeight(1);
                }
                else
                        PROGRAM.current_ticks++;
        }
        asm     sti
}




/* VESA helping functions */

int DetectVESA()
{
        _ES=FP_SEG(&VESA);
        _DI=FP_OFF(&VESA);
        _AX=0x4f00;
        asm int 10h

        if(_AX!=0x004f || memcmp(VESA.sig,"VESA",4))       /* No VESA support */
                return 0;

        /*printf("Vesa BIOS extension found\n");
        printf("Version %d.%d\n",VESA.version >> 8,VESA.version & 0x00ff);
        printf("%s\n",VESA.oemstring);
        printf("%u kilobytes of memory on board\n",VESA.memory<<6);*/
        return 1;
}

/* Make a VESA mode list, all existing plus any custom */
void VESA_MakeModeList()
{
        int index,index2;
        CUSTOM *ListPtr;
        index=0;
        index2=0;
        ListPtr=VESAModeList;

        while(1)                        /* Add custom modes to list */
        {
                if( (ListPtr + index2)->mode == 0)
                        break;
                else
                        VESAmodes[index]=(ListPtr + index2)->mode;
                index++;
                index2++;
        }
        VESAmodes[index]=0xffff;        /* Terminate List */
}

/* Returns a information about a VESA mode */
WORD VESA_ModeInfo(unsigned int mode, WORD regtop)
{
        WORD pres_ax;
        WORD pres_cx;
        WORD pres_es;
        WORD pres_di;
        WORD reg_ax;
        WORD reg_cx;
        WORD reg_es;
        WORD reg_di;
        int index;
        CUSTOM *ListPtr;
        index=0;
        ListPtr=VESAModeList;
        while((ListPtr + index)->mode != 0 && (ListPtr + index)->mode != mode)
        {
                index++;
        }

        if( (ListPtr + index)->actual_VESAmode > 0){
                reg_cx = (ListPtr + index)->actual_VESAmode;
        }else{
                return 0xFFFF;          /* error */
        }

        /* Get information about the incoming VESA mode */
        /* and fills in our VESAbuffer block */
        reg_es = FP_SEG(&VESAbuffer);
        reg_di = FP_OFF(&VESAbuffer);
        reg_ax = 0x4f01;

        pres_ax = get_reg(regtop, ofsAX);
        pres_cx = get_reg(regtop, ofsCX);
        pres_es = get_reg(regtop, ofsES);
        pres_di = get_reg(regtop, ofsDI);

        set_reg(regtop, ofsAX, reg_ax);
        set_reg(regtop, ofsCX, reg_cx);
        set_reg(regtop, ofsES, reg_es);
        set_reg(regtop, ofsDI, reg_di);

        generic_oldint10h_for_c_code(regtop);

        set_reg(regtop, ofsAX, pres_ax);
        set_reg(regtop, ofsCX, pres_cx);
        set_reg(regtop, ofsES, pres_es);
        set_reg(regtop, ofsDI, pres_di);

        /* Update to new custom mode spec */
        if((ListPtr + index)->mode > 0)
        {
                VESAbuffer.bytesperline =
                        AdjustBytesPerLine((ListPtr + index)->bytesperline);
                VESAbuffer.width=(ListPtr + index)->width;
                VESAbuffer.height=(ListPtr + index)->height;
                VESAbuffer.bpp=(ListPtr + index)->bpp;
        }

        return (ListPtr + index)->actual_VESAmode;
}



#ifdef SUPPORT_DBCS

void DBCS_get_vector(void)
{
    asm {
        pusha
        push    ds
        xor     ax, ax
        mov     ds, ax
        mov     si, ax
        mov     ax, 6300h
        int     21h
        mov     bx, ds
        pop     ds
        mov     [byte ptr dbcs_supported], al
        mov     [word ptr dbcs_ds], bx
        mov     [word ptr dbcs_si], si
        popa
    }
}

#endif


int int10h_sub(WORD regtop)
{
    WORD d0;
    WORD d1;
    WORD d2;
    DWORD d3;

    d0 = get_reg(regtop, ofsAX);

    if(PROGRAM.enabled == 0) /* TSR not active */
        return 0;

    if((PROGRAM.use_pm_interface != 0) && (d0 == 0x4F0A)) /* GET Protect Mode Interlace */
    {
        d1 = get_reg(regtop, ofsBX);
        if( (d1 & 0xFF) == 0x00){
            set_reg(regtop, ofsDI, get_pm_table_offset());
            set_reg(regtop, ofsES, get_pm_table_segment());
            set_reg(regtop, ofsCX, get_pm_table_length());
            set_reg(regtop, ofsAX, 0x004F);
        }else{
            set_reg(regtop, ofsAX, 0x014F);
        }
        return 1;
    }

    if(d0 == 0x4F00)    /* VESA SuperVGA BIOS - GET SuperVGA INFORMATION */
    {
        generic_oldint10h_for_c_code(regtop);

        _DX = (PROGRAM.membytes >> 10) >> 6;
        _AX = get_reg(regtop, ofsES);
        _ES = _AX;
        _DI = get_reg(regtop, ofsDI);
        asm     mov     [es:di + 012h], dx     /* memory emulation */

        VESA_MakeModeList();

        _AX = get_reg(regtop, ofsES);
        _ES = _AX;
        _DI = get_reg(regtop, ofsDI);
        asm     mov     [es:di + 10h], ds
        asm     mov     ax, offset VESAmodes
        asm     mov     [es:di + 0eh], ax

        return 1;
    }

    if(d0 == 0x4F01)    /* GET SuperVGA MODE INFORMATION */
    {
        d1 = get_reg(regtop, ofsCX) & 0x01FF;
        if( VESA_ModeInfo(d1, regtop) == 0xFFFF ){
            set_reg(regtop, ofsAX, 0x014F);
            return 1;
        }

        /* Code to copy from our buffer to there buffer */
        _ES = get_reg(regtop, ofsES);
        _DI = get_reg(regtop, ofsDI);
        asm mov     si, offset VESAbuffer
        asm mov     cx, 256
copyloop:
        asm mov     al, [ds:si]
        asm mov     [es:di], al
        asm add     si, 1
        asm add     di, 1
        asm dec     cx
        asm jnz     copyloop

        asm sub     di, 256             /* Restore DI to correct position */
        asm mov     al, 127
        asm mov     [es:di + 1dh], al   /* Max out on pages for increased compatability */

        set_reg(regtop, ofsAX, 0x004F);
        return 1;
    }

    if(d0 == 0x4F02)    /* SET SuperVGA VIDEO MODE */
    {
        d1  = get_reg(regtop, ofsBX);
        d2  = VESA_ModeInfo(d1 & 0x01FF, regtop);       /* d2 = actual vesa mode */
        if(d2 == 0xFFFF){
            set_reg(regtop, ofsAX, 0x014F);
            return 1;
        }
        d1  = d1 & ~0x01FF;
        d1 |= d2 &  0x01FF;
        set_reg(regtop, ofsBX, d1);
        set_reg(regtop, ofsAX, 0x4F02);
        generic_oldint10h_for_c_code(regtop);
        if( get_reg(regtop, ofsAX) != 0x004F )
            return 1;

        /* CALL YOUR CARD 15khz section here */
#ifndef debug
        SetSVGA(VESAbuffer.width, VESAbuffer.height, PROGRAM.center_x, PROGRAM.center_y, PROGRAM.h_sync_delay, PROGRAM.h_disp_delay);
#endif
        /* The correct scanbyte/width and so on if it was called in its native mode */
        PROGRAM.bitsperpixel = VESAbuffer.bpp;
        if(PROGRAM.bitsperpixel == 15){
            PROGRAM.bitsperpixel = 16;
        }
        PROGRAM.scanbyte  = (VESAbuffer.width / 8) * PROGRAM.bitsperpixel;
        PROGRAM.scanwidth = VESAbuffer.width;
        SetBytesPerLine(PROGRAM.scanbyte);

        /* Workout total number of scanlines available */
        PROGRAM.scantotal=(PROGRAM.membytes/(unsigned long)PROGRAM.scanbyte);

        return 1;
    }

    if(d0 == 0x4F06)    /* GET/SET LOGICAL SCAN LINE LENGTH */
    {
        d1 = get_reg(regtop, ofsBX) & 0x00FF;
        if(d1 == 0x00 || d1 == 0x02){           /* SET */
            d2 = get_reg(regtop, ofsCX);
            if(d1 == 0x00){
                PROGRAM.scanbyte  = CalcBytesPerLine(d2, PROGRAM.bitsperpixel);
            }else{
                PROGRAM.scanbyte  = d2;
            }
            PROGRAM.scanwidth = CalcPixelsPerLine(PROGRAM.scanbyte,
                                                 PROGRAM.bitsperpixel);
            PROGRAM.scantotal = PROGRAM.membytes / PROGRAM.scanbyte;
            SetBytesPerLine(PROGRAM.scanbyte);
            set_reg(regtop, ofsBX, PROGRAM.scanbyte);
            set_reg(regtop, ofsCX, PROGRAM.scanwidth);
            set_reg(regtop, ofsDX, PROGRAM.scantotal);
            set_reg(regtop, ofsAX, 0x004F);
        }else if(_BL == 0x01){
            set_reg(regtop, ofsBX, PROGRAM.scanbyte);
            set_reg(regtop, ofsCX, PROGRAM.scanwidth);
            set_reg(regtop, ofsDX, PROGRAM.scantotal);
            set_reg(regtop, ofsAX, 0x004F);
        }else if(_BL == 0x03){
            d2 = GetMaxBytesPerLine();
            set_reg(regtop, ofsCX, CalcPixelsPerLine(d2, PROGRAM.bitsperpixel));
            set_reg(regtop, ofsBX, d2);
            set_reg(regtop, ofsAX, 0x004F);
        }else{
            set_reg(regtop, ofsAX, 0x014F);
        }
        return 1;
    }

    if(d0 == 0x4F07)    /* GET/SET Display Start */
    {
        d1 = get_reg(regtop, ofsBX) & 0x00FF;
        if(d1 == 0x00 || d1 == 0x80){           /* SET */
            d2  = get_reg(regtop, ofsDX);
            d3  = (DWORD)d2 * PROGRAM.scanbyte;
            d2  = get_reg(regtop, ofsCX);
            d3 += (DWORD)CalcBytesPerLine(d2, PROGRAM.bitsperpixel);
            SetScreenStartAddress(d3);
            if(d1 == 0x80) VsyncVga();
            set_reg(regtop, ofsAX, 0x004F);
        }else if(_BL == 0x01){
            d3 = GetScreenStartAddress();
            d2 = d3 / PROGRAM.scanbyte;
            set_reg(regtop, ofsDX, d2);
            d2 = CalcPixelsPerLine(d3 % PROGRAM.scanbyte, PROGRAM.bitsperpixel);
            set_reg(regtop, ofsCX, d2);
            set_reg(regtop, ofsAX, 0x004F);
        }else{
            set_reg(regtop, ofsAX, 0x014F);
        }
        return 1;
    }

#ifndef STDMON
#ifndef debug
    if(d0 <= 0x13)  /* Standard VGA modes */
    {
        generic_oldint10h_for_c_code(regtop);
        switch(_AX)
        {
            case 0x13:
                SetVGA(640, 200, PROGRAM.center_x, PROGRAM.center_y, PROGRAM.h_sync_delay, PROGRAM.h_disp_delay);
                SetTextHeight(1);
                if(PROGRAM.max_ticks != 0)
                        PROGRAM.current_ticks=1;    /* Could be a modex mode so re-program after a delay */
                break;

            case 0x04:
            case 0x05:
                SetVGA(320, 200, PROGRAM.center_x, PROGRAM.center_y, PROGRAM.h_sync_delay, PROGRAM.h_disp_delay);
                SetTextHeight(2);
                break;

            case 0x0D:
                SetVGA(320, 200, PROGRAM.center_x, PROGRAM.center_y, PROGRAM.h_sync_delay, PROGRAM.h_disp_delay);
                SetTextHeight(1);
                break;

            case 0x06:
                SetVGA(640, 200, PROGRAM.center_x, PROGRAM.center_y, PROGRAM.h_sync_delay, PROGRAM.h_disp_delay);
                SetTextHeight(2);
                break;

            case 0x0E:
                SetVGA(640, 200, PROGRAM.center_x, PROGRAM.center_y, PROGRAM.h_sync_delay, PROGRAM.h_disp_delay);
                SetTextHeight(1);
                break;

            case 0x0F:
            case 0x10:
            case 0x11:
            case 0x12:
                SetVGA(640, 480, PROGRAM.center_x, PROGRAM.center_y, PROGRAM.h_sync_delay, PROGRAM.h_disp_delay);
                SetTextHeight(1);
                break;

            case 0x03:
#ifdef SUPPORT_DBCS
                if(PROGRAM.use_dbcs != 0){
                    if(dbcs_supported == 0){
                        asm {
                            push    es
                            push    si
                            mov     es, [word ptr dbcs_ds]
                            mov     si, [word ptr dbcs_si]
                            mov     ax, [word ptr es:si]
                            pop     si
                            pop     es
                        }
                        if(_AX != 0){
                            SetVGA(640, 476, PROGRAM.center_x, PROGRAM.center_y, PROGRAM.h_sync_delay, PROGRAM.h_disp_delay);
                            SetTextHeight(1);
                            break;
                        }
                    }
                }
#endif
                /*  fall through  */
            case 0x02:
            case 0x07:
                SetVGAText80x25(PROGRAM.center_x, PROGRAM.center_y, PROGRAM.h_sync_delay, PROGRAM.h_disp_delay);
                d1 = get_reg(regtop, ofsAX);
                d2 = get_reg(regtop, ofsBX);
                set_reg(regtop, ofsAX, 0x1102);
                set_reg(regtop, ofsBX, 0x0000);
                generic_oldint10h_for_c_code(regtop);
                set_reg(regtop, ofsAX, d1);
                set_reg(regtop, ofsBX, d2);
                break;

            case 0x00:
            case 0x01:
                SetVGAText40x25(PROGRAM.center_x, PROGRAM.center_y, PROGRAM.h_sync_delay, PROGRAM.h_disp_delay);
                d1 = get_reg(regtop, ofsAX);
                d2 = get_reg(regtop, ofsBX);
                set_reg(regtop, ofsAX, 0x1102);
                set_reg(regtop, ofsBX, 0x0000);
                generic_oldint10h_for_c_code(regtop);
                set_reg(regtop, ofsAX, d1);
                set_reg(regtop, ofsBX, d2);
                break;
        }
        return 1;
    }
#endif
#endif

    generic_oldint10h_for_c_code(regtop);

    return 1;
}
